#
# This code would normally be auto-generated.
#
# For the purpose of prototyping and testing, we manually create this file for now,
# until we have the auto-generation pipeline.
#
# TODO: current limitations of the manual prototype:
#   - Operations with :Attribute input parameters are only added to heco types for now.
#     The operation would actually also need to be added to all other Attributes to
#     support y + Secret(x) and not only Secret(x) + y.
#       --> Use overloads for this!

from dataclasses import dataclass
from typing import Generic, List, TypeVar, Union

from xdsl_frontend.dialects.builtin import IndexType, IntegerAttr, i32
from xdsl.ir import Attribute
import heco.dialects.fhe as orig_fhe

V = TypeVar("V", bound=Attribute, covariant=True)


################################################################################################
#                                             Types                                            #
################################################################################################

@dataclass
class SecretType(Generic[V], Attribute):
    """
    Secret.

    A `Type` for `Secret`s (high-level abstraction of `Ciphertext`s encrypting
    single scalars)

    name: "secret_type"

    Parameters:
        - plaintext_type: ParameterDef[Attribute]
    """

    # TODO: @Alex, wouldn't it make sense to make this more restrictive?
    #   Can Attribute only be of type V? (the plaintext type of SecretType)
    def __add__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Add(self, other)

    def __sub__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Sub(self, other)

    def __mul__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Mul(self, other)


@dataclass
class BatchedSecretType(Generic[V], Attribute):
    """
    Batched Secret.

    A Type for Batched `Secret`s (high-level abstraction of `Ciphertext`s
    encrypting vectors of values)

    name: "batched_secret_type"

    Parameters:
        - plaintext_type: ParameterDef[Attribute]
    """

    # TODO: @Alex, wouldn't it make sense to make this more restrictive?
    #   Can Attribute only be of type V? (the plaintext type of SecretType)
    def __add__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Add(self, other)

    def __sub__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Sub(self, other)

    def __mul__(self: 'SecretType[V]', other: Attribute) -> 'SecretType[V]':
        return Mul(self, other)

    def __getitem__(self, idx: IntegerAttr[IndexType]) -> SecretType[V]:
        return Extract(self, idx)

################################################################################################
#                                        Constraints                                           #
################################################################################################


# XXX: we need to find a way to auto-generate these. Maybe we could register this over AnyOf.
AnySecretType = Union[SecretType, BatchedSecretType]

_SecretTyp = TypeVar("_SecretTyp", bound=SecretType, covariant=True)


################################################################################################
#                                        Attributes                                            #
################################################################################################

@dataclass
class SecretAttr(SecretType[_SecretTyp]):
    """
    name: "secret"

    Parameters:
        - plaintext: ParameterDef[Attribute]
        - typ: ParameterDef[_SecretTyp]
    """
    plaintext: Attribute

    @staticmethod
    def _default_type_from_args(args: List[Attribute]) -> Attribute:
        return orig_fhe.SecretType([args[0].typ])


################################################################################################
#                                          Operation                                           #
################################################################################################

def Mul(*inputs: Attribute) -> AnySecretType:
    """
    Multiplication.

    name: "fhe.mul"

    Inputs:
        - inputs = VarOperandDef(AnyAttr())

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Mul
    return Mul  # type: ignore


def Add(*inputs: Attribute) -> AnySecretType:
    """
    Addition.

    name: "fhe.add"

    Inputs:
        - inputs = VarOperandDef(AnyAttr())

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Add
    return Add  # type: ignore


def Sub(*inputs: Attribute) -> AnySecretType:
    """
    Subtraction.

    name: "fhe.sub"

    Inputs:
        - inputs = VarOperandDef(AnyAttr())

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Sub
    return Sub  # type: ignore

# TODO: no SI32Attr exists yet in xDSL, so we use i32 for input1 instead.


def Rotate(input0: BatchedSecretType[V], input1: i32) -> BatchedSecretType[V]:
    """
    Rotate.

    name: "fhe.rotate"

    Inputs:
        - input0 = OperandDef(BatchedSecretType)
        - input1 = OperandDef(i32)

    Returns:
        - output = ResultDef(BatchedSecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Rotate
    return Rotate  # type: ignore


def Relineraization(inputs: AnySecretType) -> AnySecretType:
    """
    Relineraization.

    name: "fhe.relinearize"

    Inputs:
        - inputs = OperandDef(AnySecretType)

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Relineraization
    return Relineraization  # type: ignore


def Constant(inputs: Attribute) -> AnySecretType:
    """
    FHE constant.

    Cast a value into a `Secret` containing the same type.

    name: "fhe.constant"

    Inputs:
        - inputs = OperandDef(AnyAttr())

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Constant
    return Constant  # type: ignore


def Extract(input0: BatchedSecretType[V], input1: IntegerAttr[IndexType]) -> SecretType[V]:
    """
    Element extraction operation (actually executing this under FHE is highly inefficient).

    name: "fhe.extract"

    Inputs:
        - input0 = OperandDef(BatchedSecretType)
        - input1 = OperandDef(IndexType)

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Extract
    return Extract  # type: ignore


def Insert(input0: BatchedSecretType[V], input1: SecretAttr[V],
           input2: IntegerAttr[IndexType]) -> BatchedSecretType[V]:
    """
    Element insertion operation (actually executing this under FHE is highly inefficient).

    name: "fhe.insert"

    Inputs:
        - input0 = OperandDef(BatchedSecretType)
        - input1 = OperandDef(SecretAttr)
        - input2 = OperandDef(IndexType)

    Returns:
        - output = ResultDef(AnySecretType)
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Insert
    return Insert  # type: ignore

# XXX: CombineOp omitted for now


def Materialize(input: Attribute) -> Attribute:
    """
    No-op operation used to preserve consistency of type system during type conversion.

    name: "fhe.materialize"

    Inputs:
        - input = OperandDef(AnyAttr())

    Returns:
        - output = ResultDef(AnyAttr())
    """
    # TODO: this is a hack to get the correct xDSL operation during the frontend translation.
    #   We should really find a nicer version, because here the return type is wrong.
    #   This only works because the frontend program is never evaluated, so for static
    #   type checking we only use the function signature.
    from heco.dialects.fhe import Materialize
    return Materialize  # type: ignore



from dataclasses import dataclass

from xdsl.irdl import *
from xdsl.ir import *
from xdsl.dialects.builtin import *
from xdsl.printer import Printer


@dataclass
class FHE:
    """
    FHE Dialect.

    This dialect represents a common abstraction for FHE operations.
    """
    ctx: MLContext

    def __post_init__(self):
        self.ctx.register_attr(SecretType)
        self.ctx.register_attr(BatchedSecretType)

        self.ctx.register_op(Mul)
        self.ctx.register_op(Add)
        self.ctx.register_op(Sub)
        self.ctx.register_op(Rotate)
        self.ctx.register_op(Relineraization)
        self.ctx.register_op(Constant)
        self.ctx.register_op(Extract)
        self.ctx.register_op(Insert)
        self.ctx.register_op(Materialize)


#===----------------------------------------------------------------------===#
# FHE print helpers.
#===----------------------------------------------------------------------===#

class PrettyPrinter:

    def __init__(self, printer: Printer) -> None:
        self.printer = printer

    def _print_type(self, value: SSAValue) -> None:
        self.printer.print(" : ")
        self.printer.print_attribute(value.typ)

    def _print_types(self, values: List[SSAValue]) -> None:
        for value in values[:-1]:
            self.printer.print_attribute(value.typ)
            self.printer.print(", ")
        self.printer.print_attribute(values[-1].typ)

    def _print_signature(self, operation: Operation) -> None:
        self.printer.print(" : (")
        self._print_types(operation.operands)
        self.printer.print(") -> ")
        self._print_types(operation.results)

    def _print_operands(self, operation: Operation) -> None:
        operands = operation.operands
        self.printer.print("(")
        for op in operands[:-1]:
            self.printer.print(op, ", ")
        self.printer.print(operands[-1], ")")

    def _print_attributes(self, operation: Operation) -> None:
        self.printer.print(" ", operation.attributes)

    def print_binop(self, operation: Operation) -> None:
        self._print_operands(operation)
        self._print_attributes(operation)
        self._print_signature(operation)

    def _print_vector(self, vector: ParametrizedAttribute, idx: IndexType) -> None:
        self.printer.print(vector, "[", idx, "]")

    def print_vector_extract(self, operation: Operation) -> None:
        assert(len(operation.operands) == 2 and len(operation.results) == 1)
        self._print_vector(*operation.operands)
        self._print_attributes(operation)
        self._print_type(operation.results[0])

    def print_vector_insert(self, operation: Operation) -> None:
        assert(len(operation.operands) == 3 and len(operation.results) == 1)
        vector, elem, idx = operation.operands
        self.printer.print(elem, " into ")
        self._print_vector(vector, idx)
        self._print_attributes(operation)
        self._print_type(operation.results[0])

    def print_constant(self, operation: Operation) -> None:
        assert(len(operation.operands) == 1)
        self.printer.print_attribute(operation)
        self.printer.print(" ", operation.operands[0])

    def print_rotate(self, operation: Operation) -> None:
        # `(` $x `)` `by` $i attr-dict `:` type($x)
        assert(len(operation.operands) == 1)
        self._print_operands(operation)
        self.printer.print(" by ", operation.operands[0], " ")
        self._print_attributes(operation)
        self._print_type(operation.operands[0])

#===----------------------------------------------------------------------===#
# Other helpers.
#===----------------------------------------------------------------------===#


def _derive_result_type(value: Union[Operation, SSAValue]) -> Attribute:
    if isinstance(value, Operation):
        if len(value.results) != 1:
            raise Exception(
                f"Expected exactly one result, got {len(value.results)}")

        result = value.results[0]
        if not hasattr(result, "typ"):
            raise Exception(
                f"Attribute {result} does not have operand 'typ' storing its type.")

        return result.typ
    elif isinstance(value, SSAValue):
        return value.typ
    else:
        raise Exception(
            f"Cannot derive value from type {type(value)}; {value}.")


def _unwrap_type(attr: Attribute, recursive=True) -> Attribute:
    if isinstance(attr, SecretType):
        return attr.plaintext_type
    elif isinstance(attr, BatchedSecretType):
        return attr.plaintext_type
    elif isinstance(attr, TensorType):
        if recursive:
            return _unwrap_type(attr.element_type)
        else:
            return attr.element_type
    else:
        raise Exception(
            f"Cannot access wrapped type of {attr} (of type {type(attr)}).")

#===----------------------------------------------------------------------===#
# FHE type definitions.
#===----------------------------------------------------------------------===#


@irdl_attr_definition
class SecretType(ParametrizedAttribute):
    """
    Secret.

    A `Type` for `Secrets` (high-level abstraction of `Ciphertexts` encrypting
    single scalars)
    """
    name: str = "fhe.secret_type"
    plaintext_type: ParameterDef[Attribute]

    def get_carried_type(self) -> Attribute:
        return self.plaintext_type

    @staticmethod
    @builder
    def from_plaintext_type(typ: Attribute) -> 'SecretType':
        return SecretType([typ])

    # TODO: custom printing?
    # "`<` $plaintextType `>`";


@irdl_attr_definition
class BatchedSecretType(ParametrizedAttribute):
    """
    Batched Secret.

    A Type for Batched `Secrets` (high-level abstraction of `Ciphertexts`
    encrypting vectors of values)
    """

    name: str = "fhe.batched_secret_type"
    plaintext_type: ParameterDef[Attribute]

    # TODO: how can we add inlined C++ code?
    # let extraClassDeclaration = [{
    #         SecretType getCorrespondingSecretType() const;
    #         /// Sets size = -1
    #         static BatchedSecretType get(::mlir::MLIRContext *context, ::mlir::Type plaintextType);
    # }];

    @staticmethod
    @builder
    def from_plaintext_type(typ: Attribute) -> 'BatchedSecretType':
        return BatchedSecretType([typ])

    # TODO: custom printing
    # "`<` $size `x` $plaintextType `>`"


AnySecretType: AttrConstraint = AnyOf([SecretType, BatchedSecretType])

_SecretTyp = TypeVar("_SecretTyp", bound=SecretType, covariant=True)


@irdl_attr_definition
class SecretAttr(Generic[_SecretTyp], ParametrizedAttribute):
    name = "fhe.secret"
    plaintext: ParameterDef[Attribute]
    typ: ParameterDef[_SecretTyp]

    @staticmethod
    @builder
    def from_plaintext_and_type(plaintext: Attribute, typ: _SecretTyp) -> 'SecretAttr[_SecretTyp]':
        return SecretAttr([plaintext, typ])


#===----------------------------------------------------------------------===#
# FHE Operations
#===----------------------------------------------------------------------===#


@irdl_op_definition
class Mul(Operation):
    """Multiplication."""

    name: str = "fhe.mul"

    # TODO: how can we specify traits?
    # Commutative, DeclareOpInterfaceMethods<InferTypeOpInterface>, NoSideEffect

    arguments = VarOperandDef(AnyAttr())

    output = ResultDef(AnySecretType)

    # TODO: what is "hasFolder", do we need it, and how can we specify it?
    # hasFolder = 1

    @staticmethod
    def get(*operands: Operation | SSAValue,
            result_type: AnySecretType) -> 'Mul':
        return Mul.build(operands=[list(operands)], result_types=[result_type])

    def print(self, printer: Printer) -> None:
        pp = PrettyPrinter(printer)
        pp.print_binop(self)


@irdl_op_definition
class Add(Operation):
    """Addition."""

    name: str = "fhe.add"

    # TODO: how can we specify traits?
    # Commutative, DeclareOpInterfaceMethods<InferTypeOpInterface>, NoSideEffect

    arguments = VarOperandDef(AnyAttr())

    output = ResultDef(AnySecretType)

    # TODO: what is "hasFolder", do we need it, and how can we specify it?
    # hasFolder = 1

    @staticmethod
    def get(*operands: Operation | SSAValue, result_type: AnySecretType) -> 'Add':
        return Add.build(operands=[list(operands)], result_types=[result_type])

    def print(self, printer: Printer) -> None:
        pp = PrettyPrinter(printer)
        pp.print_binop(self)


@irdl_op_definition
class Sub(Operation):
    """Subtraction."""

    name: str = "fhe.sub"

    # TODO: how can we specify traits?
    # DeclareOpInterfaceMethods<InferTypeOpInterface>, NoSideEffect

    arguments = VarOperandDef(AnyAttr())

    output = ResultDef(AnySecretType)

    # TODO: what is "hasFolder", do we need it, and how can we specify it?
    # hasFolder = 1

    @staticmethod
    def get(*operands: Operation | SSAValue,
            result_type: AnySecretType) -> 'Sub':
        return Sub.build(operands=[list(operands)], result_types=[result_type])

    def print(self, printer: Printer) -> None:
        pp = PrettyPrinter(printer)
        pp.print_binop(self)


@irdl_op_definition
class Rotate(Operation):
    """Rotate."""

    name: str = "fhe.rotate"

    # TODO: how can we specify traits?
    # AllTypesMatch<["x","output"]>, NoSideEffect

    input0 = OperandDef(BatchedSecretType)
    # TODO: no SI32Attr exists yet in xDSL. We may want to add a `_verify` check
    #   that this value is positive.
    input1 = OperandDef(i32)

    output = ResultDef(BatchedSecretType)

    # TODO: what is "hasFolder", do we need it, and how can we specify it?
    # hasFolder = 1

    @staticmethod
    def get(vector: Union[Operation, SSAValue],
            shift: Union[Operation, SSAValue]) -> 'Relineraization':
        result_type = _derive_result_type(vector)
        return Relineraization.build(operands=[vector, shift],
                                     result_types=[result_type])

    def print(self, printer: Printer) -> None:
        printer.print(" ")  # XXX: we cannot remove the operation name
        pp = PrettyPrinter(printer)
        pp.print_rotate(self)


@irdl_op_definition
class Relineraization(Operation):
    """Relinearization."""
    name: str = "fhe.relinearize"

    # TODO: how can we specify traits?
    # AllTypesMatch<["x", "output"], NoSideEffect

    inputs = OperandDef(AnySecretType)
    output = ResultDef(AnySecretType)

    @staticmethod
    def get(input: Union[Operation, SSAValue],
            typ: AnySecretType) -> 'Relineraization':
        return Relineraization.build(operands=[input], result_types=[typ])


@irdl_op_definition
class Constant(Operation):
    """
    FHE constant.

    Cast a value into a secret containing the same type.
    """
    # TODO: what happens in TableGen when you use the same variable twice?
    # Constat defines summary twice.
    name: str = "fhe.constant"

    # TODO: how can we specify traits?
    # ConstantLike, InferTypeOpInterface,
    # DeclareOpInterfaceMethods<InferTypeOpInterface>,
    # DeclareOpInterfaceMethods<OpAsmOpInterface, ["getAsmResultNames"],
    # NoSideEffect

    inputs = OperandDef(AnyAttr())

    output = ResultDef(AnySecretType)

    @staticmethod
    def get(value: Union[Operation, SSAValue],
            typ: AnySecretType) -> 'Constant':
        return Constant.build(operands=[value], result_types=[typ])

    def print(self, printer: Printer) -> None:
        printer.print(" ")  # XXX: we cannot remove the operation name
        pp = PrettyPrinter(printer)
        pp.print_constant(self)


@irdl_op_definition
class Extract(Operation):
    """
    Element extraction operation (actually executing this under FHE is highly inefficient).
    """

    # TODO: what happens in TableGen when you use the same variable twice?
    # Constant defines summary twice.

    name: str = "fhe.extract"

    # TODO: how can we specify traits?
    # TypesMatchWith<"result type matches element type of tensor",
    #                "vector", "result",
    #                "$_self.cast<BatchedSecretType>().getCorrespondingSecretType()">
    #                NoSideEffect

    input0 = OperandDef(BatchedSecretType)
    input1 = OperandDef(IntegerAttr)

    output = ResultDef(SecretType)

    @staticmethod
    def get(vector: Union[Operation, SSAValue],
            idx: Union[Operation, SSAValue]) -> 'Extract':
        result_type = SecretType.from_plaintext_type(
            _unwrap_type(_derive_result_type(vector)))
        return Extract.build(operands=[vector, idx], result_types=[result_type])

    def print(self, printer: Printer) -> None:
        printer.print(" ")  # XXX: we cannot remove the operation name
        pp = PrettyPrinter(printer)
        pp.print_vector_extract(self)


@irdl_op_definition
class Insert(Operation):
    """
    Element insertion operation (actually executing this under FHE is highly inefficient).
    """

    name: str = "fhe.insert"

    # TODO: how can we specify traits?
    # AllTypesMatch<["dest", "result"]>,
    # TypesMatchWith<"scalar type matches element type of dest",
    #                "dest", "scalar",
    #                "$_self.cast<BatchedSecretType>().getCorrespondingSecretType()">
    # NoSideEffect

    input0 = OperandDef(BatchedSecretType)
    input1 = OperandDef(SecretAttr)
    input2 = OperandDef(IntegerAttr)

    output = ResultDef(BatchedSecretType)

    # TODO: what is "hasCanonicalizeMethod", do we need it, and how can we specify it?
    # hasCanonicalizeMethod = 1

    @staticmethod
    def get(vector: Union[Operation, SSAValue],
            elem: Union[Operation, SSAValue],
            idx: Union[Operation, SSAValue]) -> 'Insert':

        result_type = BatchedSecretType.from_plaintext_type(
            _unwrap_type(_derive_result_type(elem)))
        return Insert.build(operands=[vector, elem, idx],
                            result_types=[result_type])

    def print(self, printer: Printer) -> None:
        printer.print(" ")  # XXX: we cannot remove the operation name
        pp = PrettyPrinter(printer)
        pp.print_vector_insert(self)


# XXX: CombineOp omitted for now


@irdl_op_definition
class Materialize(Operation):
    """
    No-op operation used to preserve consistency of type system during type conversion.
    """

    name: str = "fhe.materialize"

    # TODO: how can we specify traits?
    # TypesMatchWith<"result type matches element type of tensor",
    #                "vector", "result",
    #                "$_self.cast<BatchedSecretType>().getCorrespondingSecretType()">
    #                NoSideEffect

    input = OperandDef(AnyAttr())

    output = ResultDef(AnyAttr())

    # TODO: what is "hasCanonicalizeMethod" and "hasFolder" and do we need them?
    # hasFolder = 1
    # hasCanonicalizeMethod = 1

    @staticmethod
    def get(input: Union[Operation, SSAValue], result_type: Attribute) -> 'Materialize':
        return Materialize.build(operands=[input],
                                 result_types=[result_type])
